///////////////////////////////////////////////////////////////////////
//                                                                   //
//                      RANDOM QUANTUM CIRCUIT                       //
//                WITH WEAK CONTINUOUS MEASUREMENTS                  //
//                         (CAUCHY PROCESS)                          //
//                                                                   //
//   Needs Eigen library to compile.                                 //
//                                                                   //
//   Arguments (all optional):                                       //
//     L         system size (default 16)                            //
//     T         transmission for the Poisson kernel (default 1.0)   //
//     lambda0   weak measurement parameter / T (default 0.5)        //
//     p         probability of measurement (default 0.5)            //
//     steps     number of time steps (default 100)                  //
//     verbose   print all the steps (default 0)                     //
//                                                                   //
///////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <iostream>
#include <cmath>
#include <complex>
#include <random>
#include <chrono>
#include <iterator>
#include <algorithm>

#include "Eigen/Dense"
#include "Eigen/QR"
#include "Eigen/Eigenvalues"

using namespace std;
using namespace Eigen;

double const pi = acos(-1.);
double TMI = 0;
mt19937_64 generator;
normal_distribution<double> distribution(0.0,1.0);
normal_distribution<double> distributionUp;
normal_distribution<double> distributionDown;
uniform_real_distribution<double> uni_distribution(0.0,1.0);
Matrix<complex<double>,4,4> Id4;

// Integer exponentiation
int ipow(int base, int exp)
{
    int result = 1;
    for (;;)
    {
        if (exp & 1)
            result *= base;
        exp >>= 1;
        if (!exp)
            break;
        base *= base;
    }
    return result;
}

// Reads a spin
bool bit(int number, int position)
{
    return ((number >> position) & 0x01);
}

// Reads two consecutive spins (pbc apply)
int twobits(int number, int position, int L)
{
    if(position<L-1)
        return ((number >> position) & 0x03);
    else
        return (bit(number,0)<<1)|(bit(number,L-1));
}

// Sets spin to "up"
int setbit1(int number, int position)
{
    return (number|(1UL<<position));
}

// Sets spin to "down"
int setbit0(int number, int position)
{
    return (number&~(1UL<<position));
}

// Sets two consecutive spins (pbc apply)
int setbit(int number, int position, int whattoset, int L)
{
    int x = number;
    if(position<L-1)
        x = ((~(3UL<<position))&x)|(whattoset<<position);
    else
    switch (whattoset)
    {
        case 0:
            x=setbit0(x,position%L);
            x=setbit0(x,(position+1)%L);
            break;

        case 1:
            x=setbit1(x,position%L);
            x=setbit0(x,(position+1)%L);
            break;

        case 2:
            x=setbit0(x,position%L);
            x=setbit1(x,(position+1)%L);
            break;

        case 3:
            x=setbit1(x,position%L);
            x=setbit1(x,(position+1)%L);
            break;

        default:
            break;
    }
    return x;
}

// Weak measurement of psi
void psiMeasurement_slow(complex<double>* psi, int sizeL, double lambda, int position)
{
    // Choose x according to the distribution from weak measurement
    double sumUp = 0, sumDown = 0, x;

    for(int i = 0; i < sizeL; i++)
        if(bit(i,position))
            sumUp += norm(psi[i]);
    sumDown = 1.0 - sumUp;

    x = (uni_distribution(generator) < sumUp ? distributionUp(generator) : distributionDown(generator));

    // Apply measurement to psi
    double cUp   = 1.0/sqrt( sumUp + sumDown * exp(-4.0*x*lambda) );
    double cDown = 1.0/sqrt( sumDown + sumUp * exp(+4.0*x*lambda) );

    for(int i = 0; i < sizeL; i++)
        psi[i] *= ( bit(i,position) ? cUp : cDown );
}

// Weak measurement of psi
void psiMeasurement(complex<double>* psi, int sizeL, double lambda, int position)
{
    // Choose x according to the distribution from weak measurement
    double sumUp = 0, sumDown = 0, x;
	int sizeB = ipow(2,position), sizeA = sizeL/sizeB/2, sizeB2 = sizeB*2, temp, temp2;

    for(int i = 0; i < sizeA; i++)
	{
		temp = i*sizeB2 + sizeB;
	    for(int j = 0; j < sizeB; j++)
	        sumUp += norm(psi[temp + j]);
	        //sumUp += norm(psi[i*sizeB2 + sizeB + j]);
	}
    sumDown = 1.0 - sumUp;

    x = (uni_distribution(generator) < sumUp ? distributionUp(generator) : distributionDown(generator));

    // Apply measurement to psi
    double cUp   = 1.0/sqrt( sumUp + sumDown * exp(-4.0*x*lambda) );
    double cDown = 1.0/sqrt( sumDown + sumUp * exp(+4.0*x*lambda) );

    for(int i = 0; i < sizeA; i++)
	{
		temp = i*sizeB2;
		temp2= i*sizeB2 + sizeB;
	    for(int j = 0; j < sizeB; j++)
        	psi[temp + j] *= cDown;
        	//psi[i*sizeB2 + j] *= cDown;
	    for(int j = 0; j < sizeB; j++)
        	psi[temp2 + j] *= cUp;
        	//psi[i*sizeB2 + sizeB + j] *= cUp;
	}
}

void mutualInformation(const complex<double>* psi, int L, int piece, bool verbose, double vNentropy)
{
    chrono::steady_clock::time_point begin;
    chrono::steady_clock::time_point end;
    double entropy, eig, mutualinf = 0, S_A = 0, S_B = 0, S_C = 0, S_AB = 0, S_AC = 0, S_BC = 0, S_ABC = 0;
	int lenA = ipow(2,L/piece), lenApr = ipow(2,L-(L/piece)), lenSub = ipow(2,L/2-(L/piece)), lenHalf = ipow(2,L/2), lenSubpr = ipow(2,L/2+(L/piece)), lenAA = lenA*lenA;

	if(L%piece==0)
	{
        if(verbose) begin = chrono::steady_clock::now();
        if(verbose) cout << endl << "Setup:      " << flush;
	    SelfAdjointEigenSolver< MatrixXcd > eigensolverA(lenA);
	    SelfAdjointEigenSolver< MatrixXcd > eigensolverAB(lenA*lenA);
        MatrixXcd rhoAIaux(lenApr,lenA);
        MatrixXcd rhoAIauxlong(lenSub*lenSub,lenA*lenA);
	    MatrixXcd rhoA(lenA,lenA);
	    MatrixXcd rhoB(lenA,lenA);
	    MatrixXcd rhoAB(lenA*lenA,lenA*lenA);
        if(verbose) end   = chrono::steady_clock::now();
        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << fixed << endl;

	    // S_A (lenA, lenApr)
        if(verbose) begin = chrono::steady_clock::now();
        if(verbose) cout << "RhoA setup:      " << flush;
	    for(int b1 = 0; b1 < lenA; b1++)
	    for(int b5 = 0; b5 < lenApr; b5++)
	        rhoAIaux(b5,b1) = psi[b1*lenApr + b5];
		rhoA = rhoAIaux.adjoint()*rhoAIaux;
        if(verbose) end   = chrono::steady_clock::now();
        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << fixed << endl;

        if(verbose) begin = chrono::steady_clock::now();
        if(verbose) cout << "RhoA diag:      " << flush;
	    eigensolverA.compute(rhoA,EigenvaluesOnly);
	    entropy  = 0;
	    for(int j = 0; j < lenA; j++)
	    {
	        eig = real(eigensolverA.eigenvalues()(j));
	        if(eig>1.e-15)
	        {
	            entropy  -= eig * log(eig);
	        }
	    }
	    S_A = entropy;
        if(verbose) end   = chrono::steady_clock::now();
        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." <<S_A << fixed << endl;

	    // S_(AuB) (lenA, lenSub, lenA, lenSub)
        if(verbose) begin = chrono::steady_clock::now();
        if(verbose) cout << "RhoAB setup:      " << flush;
	    for(int b1 = 0; b1 < lenA; b1++)
	    for(int b3 = 0; b3 < lenA; b3++)
	    for(int b5 = 0; b5 < lenSub; b5++)
	    for(int b6 = 0; b6 < lenSub; b6++)
	        rhoAIauxlong(b5*lenSub+b6,b1*lenA+b3) = psi[b1*lenApr + b3*lenSub + b5*lenHalf + b6];
        rhoAB = rhoAIauxlong.adjoint()*rhoAIauxlong;
        if(verbose) end   = chrono::steady_clock::now();
        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << fixed << endl;

        if(verbose) begin = chrono::steady_clock::now();
        if(verbose) cout << "RhoAB diag:      " << flush;
	    eigensolverAB.compute(rhoAB,EigenvaluesOnly);
	    entropy  = 0;
	    for(int j = 0; j < lenAA; j++)
	    {
	        eig = real(eigensolverAB.eigenvalues()(j));
	        if(eig>1.e-15)
	        {
	            entropy  -= eig * log(eig);
	        }
	    }
	    S_AB = entropy;
        if(verbose) end   = chrono::steady_clock::now();
        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." <<S_AB<< fixed << endl;

	    // S_B (lenHalf, lenA, lenSub
        if(verbose) begin = chrono::steady_clock::now();
        if(verbose) cout << "RhoB setup:      " << flush;
	    for(int b1 = 0; b1 < lenA; b1++)
	    for(int b5 = 0; b5 < lenHalf; b5++)
	    for(int b6 = 0; b6 < lenSub; b6++)
	        rhoAIaux(b5*lenSub + b6, b1) = psi[b5*lenHalf + b1*lenSub + b6];
        rhoA = rhoAIaux.adjoint()*rhoAIaux;
        if(verbose) end   = chrono::steady_clock::now();
        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << fixed << endl;

        if(verbose) begin = chrono::steady_clock::now();
        if(verbose) cout << "RhoB diag:      " << flush;
	    eigensolverA.compute(rhoA,EigenvaluesOnly);
	    entropy  = 0;
	    for(int j = 0; j < lenA; j++)
	    {
	        eig = real(eigensolverA.eigenvalues()(j));
	        if(eig>1.e-15)
	        {
	            entropy  -= eig * log(eig);
	        }
	    }
	    S_B = entropy;
        if(verbose) end   = chrono::steady_clock::now();
        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." <<S_B<< fixed;

		mutualinf = S_A - S_AB + S_B;

		if(piece==4)
		{
		    // S_C (lenA, lenSub, lenHalf)
	        if(verbose) begin = chrono::steady_clock::now();
	        if(verbose) cout << endl << "RhoC setup:      " << flush;
		    for(int b1 = 0; b1 < lenSub; b1++)
		    for(int b5 = 0; b5 < lenA; b5++)
		    for(int b6 = 0; b6 < lenHalf; b6++)
		        rhoAIaux(b5*lenHalf + b6, b1) = psi[b5*lenApr + b1*lenHalf + b6];
	        rhoA = rhoAIaux.adjoint()*rhoAIaux;
	        if(verbose) end   = chrono::steady_clock::now();
	        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << fixed << endl;

	        if(verbose) begin = chrono::steady_clock::now();
	        if(verbose) cout << "RhoC diag:      " << flush;
		    eigensolverA.compute(rhoA,EigenvaluesOnly);
		    entropy  = 0;
		    for(int j = 0; j < lenSub; j++)
		    {
		        eig = real(eigensolverA.eigenvalues()(j));
		        if(eig>1.e-15)
		        {
		            entropy  -= eig * log(eig);
		        }
		    }
		    S_C = entropy;
	        if(verbose) end   = chrono::steady_clock::now();
	        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << S_C<< fixed << endl;

		    // S_AC (lenHalf, lenHalf)
/*	        if(verbose) begin = chrono::steady_clock::now();
	        if(verbose) cout << "RhoAC setup:      " << flush;
		    for(int b1 = 0; b1 < lenHalf; b1++)
		    for(int b5 = 0; b5 < lenHalf; b5++)
		        rhoAIauxlong(b5,b1) = psi[b1*lenHalf + b5];
			rhoAB = rhoAIauxlong.adjoint()*rhoAIauxlong;
	        if(verbose) end   = chrono::steady_clock::now();
	        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << fixed << endl;

	        if(verbose) begin = chrono::steady_clock::now();
	        if(verbose) cout << "RhoAC diag:      " << flush;
		    eigensolverAB.compute(rhoAB,EigenvaluesOnly);
		    entropy  = 0;
		    for(int j = 0; j < lenHalf; j++)
		    {
		        eig = real(eigensolverAB.eigenvalues()(j));
		        if(eig>1.e-15)
		        {
		            entropy  -= eig * log(eig);
		        }
		    }
		    S_AC = entropy;
	        if(verbose) end   = chrono::steady_clock::now();
	        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << S_AC << fixed << endl;
*/
			S_AC = vNentropy;

		    // S_BC (lenA, lenHalf, lenSub)
	        if(verbose) begin = chrono::steady_clock::now();
	        if(verbose) cout << "RhoBC setup:      " << flush;
		    for(int b1 = 0; b1 < lenHalf; b1++)
		    for(int b5 = 0; b5 < lenA; b5++)
		    for(int b6 = 0; b6 < lenSub; b6++)
		        rhoAIauxlong(b5*lenSub + b6, b1) = psi[b5*lenApr + b1*lenSub + b6];
	        rhoAB = rhoAIauxlong.adjoint()*rhoAIauxlong;
	        if(verbose) end   = chrono::steady_clock::now();
	        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << fixed << endl;

	        if(verbose) begin = chrono::steady_clock::now();
	        if(verbose) cout << "RhoBC diag:      " << flush;
		    eigensolverAB.compute(rhoAB,EigenvaluesOnly);
		    entropy  = 0;
		    for(int j = 0; j < lenHalf; j++)
		    {
		        eig = real(eigensolverAB.eigenvalues()(j));
		        if(eig>1.e-15)
		        {
		            entropy  -= eig * log(eig);
		        }
		    }
		    S_BC = entropy;
	        if(verbose) end   = chrono::steady_clock::now();
	        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << S_BC <<fixed << endl;

		    // S_ABC = S_D (lenSubpr, lenSub)
	        if(verbose) begin = chrono::steady_clock::now();
	        if(verbose) cout << "RhoABC setup:      " << flush;
		    for(int b1 = 0; b1 < lenSub; b1++)
		    for(int b5 = 0; b5 < lenSubpr; b5++)
		        rhoAIaux(b5,b1) = psi[b5*lenSub + b1];
			rhoA = rhoAIaux.adjoint()*rhoAIaux;
	        if(verbose) end   = chrono::steady_clock::now();
	        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << fixed << endl;

	        if(verbose) begin = chrono::steady_clock::now();
	        if(verbose) cout << "RhoABC diag:      " << flush;
		    eigensolverA.compute(rhoA,EigenvaluesOnly);
		    entropy  = 0;
		    for(int j = 0; j < lenSub; j++)
		    {
		        eig = real(eigensolverA.eigenvalues()(j));
		        if(eig>1.e-15)
		        {
		            entropy  -= eig * log(eig);
		        }
		    }
		    S_ABC = entropy;
	        if(verbose) end   = chrono::steady_clock::now();
	        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << S_ABC<<fixed;

			TMI = S_A + S_B + S_C - S_AB - S_AC - S_BC + S_ABC;
		}
    }
    if(verbose) cout << endl << "Mutual information 1/" << piece << ":";
    cout << "\t" << mutualinf ;//<< "\t" << mutualinf2 << "\t" << mutualinf3 ;
}

void applyU(complex<double>* psi, Matrix<complex<double>,4,4>& U, int L, int sizeL, int position)
{
	complex<double> temp, temp2, temp3, temp4;
    if(position < L - 1)
    {
        int sizeLA = ipow(2,position);
        int sizeLB = sizeL/sizeLA/4;
        int iLBpr, iLBbis;

        for(int iB = 0; iB < sizeLB; iB++)
		{
            iLBpr = iB*sizeLA*4;
            for(int iA = 0; iA < sizeLA; iA++)
			{
				iLBbis = iLBpr + iA;
                temp  = psi[iLBbis           ];
                temp2 = psi[iLBbis +   sizeLA];
                temp3 = psi[iLBbis + 2*sizeLA];
                temp4 = psi[iLBbis + 3*sizeLA];
                psi[iLBbis           ] = U(0,0)*temp + U(0,1)*temp2 + U(0,2)*temp3 + U(0,3)*temp4;
                psi[iLBbis +   sizeLA] = U(1,0)*temp + U(1,1)*temp2 + U(1,2)*temp3 + U(1,3)*temp4;
                psi[iLBbis + 2*sizeLA] = U(2,0)*temp + U(2,1)*temp2 + U(2,2)*temp3 + U(2,3)*temp4;
                psi[iLBbis + 3*sizeLA] = U(3,0)*temp + U(3,1)*temp2 + U(3,2)*temp3 + U(3,3)*temp4;
            }
        }
    }
    else
    {
        int sizeLA = sizeL/2;
        for(int p = 0; p < sizeLA; p+=2)
		{
            temp  = psi[         p    ];
            temp2 = psi[sizeLA + p    ];
            temp3 = psi[         p + 1];
            temp4 = psi[sizeLA + p + 1];
            psi[         p    ] = U(0,0)*temp + U(0,1)*temp2 + U(0,2)*temp3 + U(0,3)*temp4;
            psi[sizeLA + p    ] = U(1,0)*temp + U(1,1)*temp2 + U(1,2)*temp3 + U(1,3)*temp4;
            psi[         p + 1] = U(2,0)*temp + U(2,1)*temp2 + U(2,2)*temp3 + U(2,3)*temp4;
            psi[sizeLA + p + 1] = U(3,0)*temp + U(3,1)*temp2 + U(3,2)*temp3 + U(3,3)*temp4;
        }
    }
}

// Generate a random CUE matrix
Matrix<complex<double>,4,4> cue()
{
    double number = distribution(generator);
    Matrix<complex<double>,4,4> z;

    for(int i = 0; i < 4; i++)
        for(int j = 0; j < 4; j++)
            z(i,j) = complex<double>(distribution(generator),distribution(generator));

    HouseholderQR<Matrix<complex<double>,4,4> > qr(z.rows(),z.cols());
    qr.compute(z);
    Matrix<complex<double>,4,4> r = qr.matrixQR(); // .triangularView<Upper>() to get actual R
    Matrix<complex<double>,4,4> q = qr.householderQ();
    Matrix<complex<double>,4,4> rd;

    for(int i = 0; i < 4; i++)
        rd(i,i) = r(i,i)/abs(r(i,i));

    return (q * rd);
}

// Generate a Poisson kernel
Matrix<complex<double>,4,4> poissonKernel(double T)
{
    double R = sqrt(1.-T*T);
    return R*Id4 - T*T*(cue()-Id4*R).inverse();
}

// Generate a Gaussian kernel
Matrix<complex<double>,4,4> gaussianKernel(double mu)
{
	Matrix<complex<double>,4,4> h;
    for(int j = 0; j < 4; j++)
        for(int i = 0; i < 4; i++)
            h(i,j) = complex<double>(distribution(generator),distribution(generator))/sqrt(2.0);
	h = (h + h.adjoint().eval())*mu*mu*complex<double>(0.,0.5);

    return (Id4 - h)*(Id4 + h).inverse();
}

int main(int argc, char const *argv[])
{
    // System setup
    int    L       = 16;
    double lambda  = 0.5;
    double T       = 1.0;
    double p       = 0.5;
    int    steps   = 100;
    bool   verbose = 0;
    double lambda0;

    if(argc==5)
    {
        L       = atoi(argv[1]);
        T       = atof(argv[2]);
        lambda0 = atof(argv[3]);
        p       = atof(argv[4]);
        lambda  = lambda0*T*T;
    }
    if(argc==6)
    {
        L       = atoi(argv[1]);
        T       = atof(argv[2]);
        lambda0 = atof(argv[3]);
        p       = atof(argv[4]);
        steps   = atoi(argv[5]);
        lambda  = lambda0*T*T;
    }
    if(argc==7)
    {
        L       = atoi(argv[1]);
        T       = atof(argv[2]);
        lambda0 = atof(argv[3]);
        p       = atof(argv[4]);
        steps   = atoi(argv[5]);
        verbose = atoi(argv[6]);
        lambda  = lambda0*T*T;
    }

	generator.seed(chrono::system_clock::now().time_since_epoch().count());
//	generator.seed(10);
    distributionUp   = normal_distribution<double>(+lambda,1.0/sqrt(2.0));
    distributionDown = normal_distribution<double>(-lambda,1.0/sqrt(2.0));
    int sizeL = ipow(2,L);
    int sizeL2 = ipow(2,L/2);
    complex<double>* psi  = new complex<double>[sizeL];
    SelfAdjointEigenSolver< MatrixXcd > eigensolver(sizeL2);
    MatrixXcd rhoA(sizeL2,sizeL2);
    Vector4cd v1(0,0,0,0), v2(0,0,0,0);
    Matrix<complex<double>,4,4> U [L];
    double entropy, entropy2, entropy3, eig, xprob, sum;
    complex<double> sumc;
    chrono::steady_clock::time_point begin;
    chrono::steady_clock::time_point end;
	Id4.setIdentity();
    cout << fixed;
    cout.precision(15);
	int position_of_U = (uni_distribution(generator)<0.5) ? 0 : 1;

    // Start from the Neel state
    psi[(sizeL-1)/3] = complex<double>(1.,0.);

    // Main loop
    for(int loop = 0; loop < steps; loop++)
    {
        // Generate the U matrices
        if(verbose) cout << "Generating U matrices." << endl;
        for(int i = 0; i < L; i++)
            U[i] = gaussianKernel(T);

        // Apply U on even pairs
        if(verbose) begin = chrono::steady_clock::now();
        if(verbose) cout << "Applying U even: " << flush;
        for(int pos = position_of_U; pos < L; pos+=2)
        {
			applyU(psi, U[pos], L, sizeL, pos);
            if(verbose) cout << "." << flush;
        }
        if(verbose) cout << endl;
        if(verbose) end   = chrono::steady_clock::now();
        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << fixed << endl;

        // Apply M
        if(verbose) begin = chrono::steady_clock::now();
        if(verbose) cout << "Applying M:      " << flush;
        for(int pos = 0; pos < L; pos++)
        {
            if( uni_distribution(generator) < p )
            {
                psiMeasurement(psi, sizeL, lambda, pos);
            }
            if(verbose) cout << "." << flush;
        }
        if(verbose) cout << endl;
        if(verbose) end   = chrono::steady_clock::now();
        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << fixed << endl;

        // Apply U on even pairs
        if(verbose) begin = chrono::steady_clock::now();
        if(verbose) cout << "Applying U even: " << flush;
        for(int pos = 1-position_of_U; pos < L; pos+=2)
        {
			applyU(psi, U[pos], L, sizeL, pos);
            if(verbose) cout << "." << flush;
        }
        if(verbose) cout << endl;
        if(verbose) end   = chrono::steady_clock::now();
        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << fixed << endl;

        // Apply M
        if(verbose) begin = chrono::steady_clock::now();
        if(verbose) cout << "Applying M:      " << flush;
        for(int pos = 0; pos < L; pos++)
        {
            if( uni_distribution(generator) < p )
            {
                psiMeasurement(psi, sizeL, lambda, pos);
            }
            if(verbose) cout << "." << flush;
        }
        if(verbose) cout << endl;
        if(verbose) end   = chrono::steady_clock::now();
        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << fixed << endl;

        if( (loop+1)%((int)round(0.25/T/T/T/T))==0 ) // assumes autocorrelation time ~ 0.25
		{
	        // Calculate reduced density matrix
	        if(verbose) cout << defaultfloat;
	        if(verbose) cout << "Calculating rhoA... " << flush;
	        if(verbose) begin = chrono::steady_clock::now();
	        for(int b1 = 0; b1 < sizeL2; b1++)
	            for(int b2 = 0; b2 < sizeL2; b2++)
	                rhoA(b2,b1) = psi[b1*sizeL2 + b2];
	        rhoA = rhoA.adjoint()*rhoA;
	        if(verbose) end   = chrono::steady_clock::now();
	        if(verbose) cout << "Completed in " << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << endl;

	        // Calculate eigenvalues of reduced density matrix
	        if(verbose) cout << "Calculating rhoA eigenvalues... " << flush;
	        if(verbose) begin = chrono::steady_clock::now();
	        eigensolver.compute(rhoA,EigenvaluesOnly);
	        if(verbose) end   = chrono::steady_clock::now();
	        if(verbose) cout << "Completed in " << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << endl;

	        // Calculate entropy
	        if(verbose) cout << fixed;
	        if(verbose) cout << "Von Neumann entropy: " << flush;
	        entropy  = 0;
	        entropy2 = 0;
	        entropy3 = 0;
	        for(int j = 0; j < sizeL2; j++)
	        {
	            eig = real(eigensolver.eigenvalues()(j));
	            if(eig>1.e-15)
	            {
	                entropy  -= eig * log(eig);
	                entropy2 += eig*eig;
	                entropy3 += eig*eig*eig;
	            }
	        }
	        entropy2 = -log(entropy2);
	        entropy3 = -0.5 * log(entropy3);
	        if(!verbose) cout << defaultfloat << (loop+1)*T*T*T*T << fixed << "\t" << flush;
	        cout << entropy << "\t" << flush;
	        if(verbose) cout << endl << "2-Rényi entropy:     " << flush;
	        cout << entropy2 << "\t" << flush;
	        if(verbose) cout << endl << "3-Rényi entropy:     " << flush;
	        cout << entropy3 << flush;

	        mutualInformation(psi, L, 4, verbose, entropy);
	        mutualInformation(psi, L, 6, verbose, entropy);
	        mutualInformation(psi, L, 8, verbose, entropy);
			if(L%4==0)
			{
			    if(verbose) cout << endl << "Tripartite mutual information:";
    			cout << "\t" << TMI ;
			}
	        cout << endl;
		}

        // Check the norm of the wavefunction
        if(verbose)
        {
            cout << "Norm check:          " << flush;
            sum = 0;
            for(int i = 0; i < sizeL; i++)
                sum += norm(psi[i]);
            cout << sum << endl;
        }
		if(loop%100==0)
		{
	        if(verbose) begin = chrono::steady_clock::now();
			sum = 0;
            for(int i = 0; i < sizeL; i++)
                sum += norm(psi[i]);
			sum = sqrt(sum);
            for(int i = 0; i < sizeL; i++)
                psi[i] /= sum;
			if(verbose) cout << "Wavefunction normalised. ";
	        if(verbose) end   = chrono::steady_clock::now();
	        if(verbose) cout << "Completed in " << defaultfloat << (double)(chrono::duration_cast<chrono::milliseconds>(end - begin).count())/1000. << " s." << fixed << endl;
		}
    }

    return 0;
}
